package view;

import javafx.scene.Group;
import javafx.scene.shape.*;
import model.Board;
import model.Pawn;
import model.Piece;
import model.Position;

import java.util.ArrayList;

public class BoardView {

	public static final int BOARD_SIZE = 16, CENTERED_ORIGIN = -BOARD_SIZE*7/2;

	private TileView[] tiles; // contains all the squares/tiles of the board (visual)
	private PieceView[] pieces; // contains all the 3d pieces (visual)
	private int black_death_count, white_death_count; // graveyard counter

	public BoardView(Custom3dModel model, CustomMaterial material) {
		// creates the board squares, and position them properly
		tiles = new TileView[64];
		for(int j = 1 ; j <= 8 ; j++) {
			for (int i = 1 ; i <= 8; i++) {
				createTile(i, j);
			}
		}
		resetColor(material); // init/reset the board squares with black and white

		// creates the 3d pieces
		pieces = new PieceView[64];
		for(int n = 0 ; n < 64 ; n++) {
			pieces[n] = null;
		}
		int y;
		for(int color = Piece.WHITE; color == Piece.WHITE || color == Piece.BLACK ; color++) {
			// placement of pawns
			if(color == Piece.WHITE) { y = 2; } else { y = 7; }
			for(int x = Position.a ; x <= Position.h ; x++) {
				createPiece(x, y, model.get(Custom3dModel.PAWN), color, material);
			}
			// placement of rook, knight, bishop, queen, king
			if(color == Piece.WHITE) { y = 1; } else { y = 8; }
			createPiece(Position.a, y, model.get(Custom3dModel.ROOK), color, material);
			createPiece(Position.b, y, model.get(Custom3dModel.KNIGHT), color, material);
			createPiece(Position.c, y, model.get(Custom3dModel.BISHOP), color, material);
			createPiece(Position.d, y, model.get(Custom3dModel.QUEEN), color, material);
			createPiece(Position.e, y, model.get(Custom3dModel.KING), color, material);
			createPiece(Position.f, y, model.get(Custom3dModel.BISHOP), color, material);
			createPiece(Position.g, y, model.get(Custom3dModel.KNIGHT), color, material);
			createPiece(Position.h, y, model.get(Custom3dModel.ROOK), color, material);
		}
		// init graveyard empty
		black_death_count = 0;
		white_death_count = 0;
	}

	public void updateAllPieces() { // updates all the pieces, visually, to the right position
		for(int j = 1 ; j <= 8 ; j++) {
			for (int i = 1; i <= 8; i++) {
				updatePiece(new Position(i, j));
			}
		}
	}

	public void updatePiece(Position position) { // updates the piece(i, j), visually, to the right position
		int id = position.getID();
		if(getPiece(id) != null) {
			getPieceObj(id).setTranslateX((position.getX()-1) * BOARD_SIZE + CENTERED_ORIGIN);
			getPieceObj(id).setTranslateZ((position.getY()-1) * BOARD_SIZE + CENTERED_ORIGIN);
		}
	}

	public void updateGraveyardPiece(PieceView piece) { // updates position (visual) of a dead piece
		int x = piece.getPosition().getX();
		int y = piece.getPosition().getY();
		if(x == Position.WHITE_GRAVEYARD) {
			piece.getObj().setTranslateX(-2 * BOARD_SIZE + CENTERED_ORIGIN);
			piece.getObj().setTranslateZ(y * BOARD_SIZE*0.65f - BOARD_SIZE*4);
		} else { // x == Position.BLACK_GRAVEYARD
			piece.getObj().setTranslateX(9 * BOARD_SIZE + CENTERED_ORIGIN);
			piece.getObj().setTranslateZ(-y * BOARD_SIZE*0.65f + BOARD_SIZE*4);
		}
		piece.getObj().setTranslateY(-1); // place at right height (Y axis) on table
	}

	public void createTile(int x, int y) {
		// creates a tile and inits the right size and position
		Box box = new Box(BOARD_SIZE, BOARD_SIZE / 2, BOARD_SIZE);
		box.setTranslateX((x-1) * BOARD_SIZE + CENTERED_ORIGIN);
		box.setTranslateZ((y-1) * BOARD_SIZE + CENTERED_ORIGIN);
		// sets the tile on the list
		setTile(new TileView(box, new Position(x, y)));
	}

	private void createPiece(int x, int y, TriangleMesh mesh, int color, CustomMaterial material) {
		// creates a piece, and inits materials, drawmode, scale, and the right height (Y axis)
		MeshView obj = new MeshView(mesh);
		obj.setDrawMode(DrawMode.FILL);
		if(color == Piece.WHITE)
			obj.setMaterial(material.get(CustomMaterial.WHITE));
		else
			obj.setMaterial(material.get(CustomMaterial.BLACK));
		int scale = 16; // scales the 3d models to a wanted size, to fit a board tile
		obj.setScaleX(scale);
		obj.setScaleY(-scale); // Y axis is down (not up) in javaFX, so we need to reverse object
		obj.setScaleZ(-scale); // we reverse Z axis too, "to keep the orientation of the 3-axis vectors", (reversing Y and Z is equivalent of 180° rotation of X axis)
		obj.setTranslateY(-5); // place at right height (Y axis) on board
		// if color is black we rotate
		if(color == Piece.BLACK) {
			obj.setScaleX(-obj.getScaleX());
			obj.setScaleZ(-obj.getScaleZ());
		}
		// sets the piece on the list
		Position position = new Position(x, y);
		setPiece(new PieceView(obj, position, color));
		// updates the right board-tile position x, y (corresponding to X axis and Z axis placements)
		updatePiece(position);
	}

	public TileView[] getTiles() {
		return tiles;
	}

	public TileView getTile(int id) {
		return tiles[id];
	}

	public Box getTileObj(int id) {
		return tiles[id].getObj();
	}

	public void setTile(TileView tile) {
		tiles[tile.getPosition().getID()] = tile;
	}

	public PieceView[] getPieces() {
		return pieces;
	}

	public PieceView getPiece(int id) {
		return pieces[id];
	}

	public MeshView getPieceObj(int id) {
		return pieces[id].getObj();
	}

	public void setPiece(PieceView piece) {
		pieces[piece.getPosition().getID()] = piece;
	}

	public void resetColor(CustomMaterial material) {
		for(int n = 0, j = 1 ; j <= 8 ; j++) {
			for (int i = 1; i <= 8; i++, n++) {
				// resets board with black and white - bottom left (first case) is black, then alternates
				if((i + j) % 2 == 0) {
					getTileObj(n).setMaterial(material.get(CustomMaterial.BOARDBLACK));
					getTile(n).setState(TileView.NORMAL);
				} else {
					getTileObj(n).setMaterial(material.get(CustomMaterial.BOARDWHITE));
					getTile(n).setState(TileView.NORMAL);
				}
			}
		}
	}

	public void setColor(CustomMaterial material, Position position, int type) {
		getTileObj(position.getID()).setMaterial(material.get(type));
		switch(type) {
			case CustomMaterial.SELECTED -> getTile(position.getID()).setState(TileView.SELECTED);
			case CustomMaterial.MOVABLE -> getTile(position.getID()).setState(TileView.MOVABLE);
			case CustomMaterial.ATTACK -> getTile(position.getID()).setState(TileView.ATTACK);
			case CustomMaterial.SPECIAL -> getTile(position.getID()).setState(TileView.SPECIAL);
			default -> getTile(position.getID()).setState(TileView.NORMAL);
		}
	}

	public void clickPiece(CustomMaterial material, PieceView clickedPiece, Board board) {
		Position position = clickedPiece.getPosition();
		board.setSelected(position);
		resetColor(material);
		setColor(material, position, CustomMaterial.SELECTED);
		ArrayList<Position> moves = board.getPiece(position).getAvailableMoves(board);
		for (Position move : moves) {
			if(board.isFree(move)) {
				setColor(material, move, CustomMaterial.MOVABLE);
			} else if(board.isEnnemy(move, position)) {
				setColor(material, move, CustomMaterial.ATTACK);
			} else {
				setColor(material, move, CustomMaterial.SPECIAL);
			}
			// promotion
			if(board.isPawn(position) && (move.getY() == 8 || move.getY() == 1)) {
				setColor(material, move, CustomMaterial.SPECIAL);
			}
		}
	}

	public void clickTile(CustomMaterial material, TileView clickedTile, Board board) {
		int state = clickedTile.getState();
		Position selected = board.getSelected();
		Position arrival = clickedTile.getPosition();
		if(state != TileView.NORMAL && state != TileView.SELECTED) {
			if(state == TileView.ATTACK) {
				killPiece(getPiece(arrival.getID()));
			}
			if(state == TileView.SPECIAL) {
				// promotion can be done while attacking too
				if(!board.isFree(arrival)) {
					killPiece(getPiece(arrival.getID()));
				}
				// promotion + rook , to implement
			}
			board.movePiece(selected, arrival); // move piece (on model)
			movePiece(selected, arrival); // move piece (on 3d view)
			resetColor(material);
		}
	}

	public void movePiece(Position origin, Position arrival) {
		PieceView piece = getPiece(origin.getID());
		piece.setPosition(arrival); // changes position to arrival
		setPiece(piece); // copies the piece from origin ID to arrival ID
		pieces[origin.getID()] = null; // removes the origin ID piece
		updatePiece(arrival); // updates 3d model position (visual)
	}

	public void killPiece(PieceView piece) { // graveyards are specified as negative x
		if(piece.getColor() == Piece.WHITE){
			white_death_count++;
			piece.setPosition(new Position(Position.WHITE_GRAVEYARD, white_death_count));
		} else {
			black_death_count++;
			piece.setPosition(new Position(Position.BLACK_GRAVEYARD, black_death_count));
		}
		updateGraveyardPiece(piece);
	}

	public void addAllToScene(Group group) {
		for(int n = 0 ; n < 64 ; n++) {
			if(getTile(n) != null) {
				group.getChildren().addAll(getTileObj(n));
			}
			if(getPiece(n) != null) {
				group.getChildren().addAll(getPieceObj(n));
			}
		}
	}
}
